<?php
/***************************************************************************
 *   Copyright (C) 2005 by Ferenc Veres   *
 *   lion@netngine.hu   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
include_once('TranslationWord.php');

/**
 * Class to manage all words for one sentence
 *
 * This class basically just stores an array of words (TranslationWord objects),
 * count of their appearances in the sentence, and can return these properties.
 * The main function of this class is to compare itself with another instance,
 * and return the contained words' match percentage for finding similar sentences.
 * It can be initialized two ways: Adding words one by one or parse a sentence.
 */

class TranslationWordList implements Iterator
{
	private $allWords;		/**< Array containing our TranslationWords objects */
	private $fakeLength;	/**< Used to fake SumCount() if not all words are actually loaded */

	private $DB;			/**< Database connection reference. */
	private $Tables;		/**< Database tables reference. */
	private $currIndex;		/**< Current item when using the internal Iterator */

	public static $WordsCache; /**< Cache for words already indexed in the database, initialized by the InitWordsCache public static method of this class. */

	/**
	* Creates a new TranslationWordList object, and parses the passed sentence if any.
	*
	* While parsing sentence, this class adds all words to the WordIndex database or finds
	* already existing ones. The class maintains an iterable list of TranslationWord objects.
	* @param $text string Text to parse and index. If empty, an object without actual wordlist is created.
	*/
	
	function __construct($text = "")
	{
		// Init DB references
		$this->DB = GetDbConn();
		$this->Tables = GetTables();

		$this->currIndex = 0;
		$this->fakeLength = 0;
		unset($this->allWords);

		// If text passed, parse it and fill the words arrays (also adds to WordIndex if necessary)
		if($text != "")
		{
			$this->SplitToWords($text);
		}
	}

	/**** Iterator interface ****/

	/**
	* Iterator interface: rewind function.
	*/
	public function rewind()
	{
		$this->currIndex = 0;
	}
	/**
	* Iterator interface: hasMore function.
	*/
	public function hasMore()
	{
		return count($this->allWords) > $this->currIndex;
	}
	/**
	* Iterator interface: key function.
	*/
	public function key()
	{
		return $this->currIndex;
	}
	/**
	* Iterator interface: current function.
	*/
	public function current()
	{
		return $this->allWords[$this->currIndex];
	}
	/**
	* Iterator interface: next function.
	*/
	public function next()
	{
		$this->currIndex++;
	}
	/**
	* Iterator interface: valid function.
	*/
	public function valid()
	{
	if(count($this->allWords) > $this->currIndex)
		{
			return true;
		}
		return false;
	}

	/* Static initialization methods */

	/**
	* Loads all already indexed word from the database to the public static $WordsCache variable.
	*
	* This function must be called before using any Translation functions. TODO: autoload on need?
	*/
	
	public static function InitWordsCache()
	{
		$Tables = GetTables();
		$DB = GetDbConn();
		
		// Fetch already indexed words and their ID
		$rs = $DB->Execute("SELECT Word, ID FROM ".$Tables["WordIndex"]);
	
		if(!$rs)
		{
			die("DB Error:". $this->DB->ErrorMsg());
		}
	
		while($oldWord = $rs->FetchRow())
		{
			// Already existing words get an ID now
			TranslationWordList::$WordsCache[$oldWord["Word"]] = $oldWord["ID"];
		}
	}

	/**** Add words and properties ****/
	
	/**
	* Add a word by it's actual text to our internal Word list.
	*
	* Adds a word which will not have ID, just text. If the word already
	* exist in this object, the number of occurences
	* will be increased by $count.
	*
	* @param $word The text of the word.
	* @param $count Occurences of the word in the current context, defaults to 1.
	* @returns Nothing.
	*/

	function AddWord($word, $count = 1)
	{
		// Get a reference to this word in allWords if already added
		$w =& $this->FindByName($word);
		if($w == null)
		{
			// First occurence(s) in this sentence, create with emtpy ID
			$w = new TranslationWord();
			$w->SetWord($word);
			$w->SetCount($count);
			$this->allWords[] = $w;
		}
		else
		{
			// Not first occurence(s), count it
			$w->SetCount($w->GetCount() + $count);
		}
	}

	/**
	* Add a word only by it's WordIndex ID to our internal Word list.
	*
	* Adds a word which will not have actual text, just ID. Later, if necessary, the
	* TranslationWord object can be extended to load it's text when ID is configured but
	* word text is not. If the word already exist in this context, the number of occurences
	* will be increased by $count.
	*
	* @param $id The ID of the word in the WordIndex database.
	* @param $count Occurences of the word in the current context, defaults to 1.
	* @returns Nothing.
	*/
	
	function AddID($id, $count = 1)
	{
		// Get a reference to this word in allWords if already added
		$w =& $this->FindByID($id);
		if($w == null)
		{
			// First occurence(s) in this sentence, create with emtpy Name
			$w = new TranslationWord();
			$w->SetID($id);
			$w->SetCount($count);
			$this->allWords[] = $w;
		}
		else
		{
			// Not first occurence(s), count it
			$w->SetCount($w->GetCount() + $count);
		}
	}


	/**
	* Set ID for an already added word.
	*
	* Sets the WordIndex ID of a word which is already added to our wordlist.
	*
	* @param $word The word string which will get the ID.
	* @param $id The ID to set to this word.
	*/
	 
	function SetIDByName($word, $id)
	{
		$w =& $this->FindByName($word);
		if($w == null)
		{
			die("WORDS: SetIDByName to a not existing word: $word, $id.\n");
		}

		$w->SetID($id);
	}
	
	/**
	* Set a fake value for the actual sentences' length, when the real lenght is
	* known but not all words are loaded.
	*
	* The SumCount function will return this value if set, otherwise it will really
	* count the occurences of the actually loaded words.
	* @param $length The fake length to set.
	*/
	
	function SetFakeLength($length)
	{
		$this->fakeLength = $length;
	}

	/**
	* Splits any string to human readable words and adds them to the internal $this->allWords array
	*
	* Also counts how many times each word appears in the sentence. Also adds all missing words
	* to the WordIndexes.
	*
	* @param $text The sentence to split to words
	* @returns Nothing.
	*/
	
	private function SplitToWords($text)
	{
		// Clear previous items
		unset($this->allWords);
		$this->SetFakeLength(0);
		
		// Remove linebreaks
		$text = str_replace("\n", " ", $text);
		$text = str_replace("\r", " ", $text);
		
		// Remove some HTML emphasis from the words
		$text = str_replace("<strong>", "", $text);
		$text = str_replace("</strong>", "", $text);
		$text = str_replace("<b>", "", $text);
		$text = str_replace("</b>", "", $text);
		$text = str_replace("<i>", "", $text);
		$text = str_replace("</i>", "", $text);
		$text = str_replace("<em>", "", $text);
		$text = str_replace("</em>", "", $text);
		$text = str_replace("&nbsp;", " ", $text);

		// Split to words
		$wordList = split('[ \.,()!\?:;=|*]', $text);
		
		// Filter non-words
		foreach($wordList as $word)
		{
			// Remove trash remained with split
			$word = trim($word, " \t\n\r:[]#'\"$+-_<>/");

			// Nonempty, nonnumbers
			if($word != "" && !is_numeric($word))
			{
				$this->AddWord(strtolower($word));
			}
		}

		// Add to word index database and get indexes for already indexed
		if($this->SumCount() != 0)
		{
			$this->AddToWordIndex();
		}
	}

	/**** Get words and properties ****/

	/**
	* Find a word by its text and return a REFERENCE to its TranslationWord object.
	*
	* You can change Word object using the returned reference, be careful.
	*
	* @param $word The word to find in the internal array of Words.
	* @returns Reference to a TranslationWord object, null if not found.
	*/
	function &FindByName($word)
	{
		if(empty($this->allWords))
		{
			$null = null;
			return $null;
		}
		
		foreach($this->allWords as $w)
		{
			if($w->GetWord() == $word)
			{
				return $w;
			}
		}

		$null = null;
		return $null;
	}

	/**
	* Find a word by its ID and return a REFERENCE to its TranslationWord object.
	*
	* You can change Word object using the returned reference, be careful.
	*
	* @param $id The ID of the word to find in the internal array of Words.
	* @returns Reference to a TranslationWord object, null if not found.
	*/
	function &FindByID($id)
	{
		if(empty($this->allWords))
		{
			$null = null;
			return $null;
		}
		
		foreach($this->allWords as $w)
		{
			if($w->GetID() == $id)
			{
				return $w;
			}
		}

		$null = null;
		return $null;
	}

	
	/**
	* Return the number of all words in this sentence.
	*
	* Counts all loaded words occurences in the internal Words array. If the SetFakeLength()
	* function was used to configure a different real length, that will be returned instead.
	*
	* @returns Total number of words in our Words array.
	*/
	
	function SumCount()
	{
		// Some sentences are not completely loaded but can fake their
		// original length
		if($this->fakeLength <> 0)
		{
			return $this->fakeLength;
		}

		// Not loaded sentences are 0 long
		if(empty($this->allWords))
		{
			return 0;
		}

		// Summarize all words occurences otherwise
		$count = 0;
		foreach($this->allWords as $w)
		{
			$count += $w->GetCount();
		}

		return $count;
	}

	/**** Compare ****/

	/**
	* Comare our Words array to the one in another similar object.
	*
	* This function comapres this and another TranslationWordList and returns a similarity percentage
	* value. The value is 100% if this and the passed TranslationWordList object contains the same
	* words and each appears the same times in both. Otherwise smaller than 100.
	*
	* @param $otherWords Another TranslationWordList object to compare our words to.
	* @returns Percentage of the match.
	*/
	
	function CompareByID($otherWords)
	{
		// Max points is the sum of all words from both sentences
		$max = $this->SumCount() + $otherWords->SumCount();
		
		$points = 0;
		
		// Loop on all our own words
		foreach($this->allWords as $w)
		{
			//This word exists in the other sentence too?
			$w2 = $otherWords->FindByID($w->GetID());
			if($w2 != null)
			{
				// Yes. common word. (each common word gets 2 points)
				
				// Same number of occurences?
				if($w2->GetCount() == $w->GetCount())
				{
					$points += $w->GetCount() * 2;
				}
				else
				{
					$points += abs($w2->GetCount() - $w->GetCount()) * 2;
				}
			}
		}

		// Return match percentage

		return floor((100/$max) * $points);
	}

	/**** Database stuff ****/

	/**
	* Adds the new words from the internal $this->allWords array to the word index table
	*
	* Newly inserted words will get an index in the $this->allWords array,
	* already existing words get their old ID from the table into the same array.
	*
	* @returns Nothing.
	*/
	
	private function AddToWordIndex()
	{
		// Get references to cached data (also inits the cache if necessary)
		
		// Add all new words to the index ($w is reference!)
		foreach($this->allWords as &$w)
		{
			// Not yet existing word?
			if(!empty(TranslationWordList::$WordsCache[$w->GetWord()]))
			{
				// Have word
				$this->SetIDByName($w->GetWord(), TranslationWordList::$WordsCache[$w->GetWord()]);
			}
			else
			{
				// Add new word to the index and get it's ID
				$ok = $this->DB->Execute("INSERT INTO ".$this->Tables["WordIndex"]." (Word) VALUES ('".fixstr($w->GetWord())."')");

				if(!$ok)
				{
					die("DB Error:". $this->DB->ErrorMsg()." Line: ".__LINE__);
				}
				if(!$this->DB->Insert_ID())
				{
					die("DB Error: Insert_ID not supported. Line: ".__LINE__);
				}
				
				// Store the new ID
				$w->SetID($this->DB->Insert_ID());
				TranslationWordList::$WordsCache[$w->GetWord()] = $w->GetID();
			}
		}

		// By now all Words have a name, ID and count for this Sentence
	}

}


?>
